-- | The base Gulik environment.
module Gulik.BaseEnv(baseEnv) where
-- Import a bunch of things, with care towards name collisions.
import Gulik.Types
import Gulik.Eval
import Data.Map as Map hiding(map)
import Control.Monad.Error

{- | Environment containing all built-in symbols.  See the
     object-language autogenerated documentation for descriptions of
     these symbols. -}
baseEnv :: Env
baseEnv = fromList envList
    where
      envList :: [(String, Value)]
      envList = [-- Math operators
                 makeBinMath  (+)           "add",
                 makeBinMath  (*)           "mul",
                 makeUnMath   (\x -> (-x))  "neg",
                 makeUnDouble (\x -> (1/x)) "inv",
                 makeDtoI     truncate      "truncate",
                 makeItoD     fromInteger   "todouble",
                 makeBinRel   (<)           "lt",

                 -- Equality
                 makeBuiltin "eq" 2 (typeCasesOp [toDoubleBinRel (==), genEq]),

                 -- Conditionals / booleans
                 makeBuiltin "if" 3 ifOp,
                 ("true",  ValueBool True ),
                 ("false", ValueBool False),

                 -- Lists
                 makeBuiltin "cons" 2 consOp,
                 makeBuiltin "car"  1 (pairOp fst),
                 makeBuiltin "cdr"  1 (pairOp snd),
                 ("nil", ValueNil),

                 -- Exceptions
                 ("error", errorBuiltin)]

      onlyInteger :: Value -> Maybe Integer
      onlyInteger (ValueInteger x) = Just x
      onlyInteger _                = Nothing

      onlyDouble  :: Value -> Maybe Double
      onlyDouble  (ValueDouble x)  = Just x
      onlyDouble  _                = Nothing

      onlyString  :: Value -> Maybe String
      onlyString  (ValueString x)  = Just x
      onlyString  _                = Nothing

      toDouble :: Value -> Maybe Double
      toDouble (ValueDouble  x) = Just x
      toDouble (ValueInteger x) = Just (fromInteger x)
      toDouble _                = Nothing

      onlyIntegerBinOp  :: (Integer -> Integer -> Integer) -> [Value] -> Maybe Value
      onlyIntegerBinOp  f vs = case mapM onlyInteger vs of
                                 Just [x, y] -> Just (ValueInteger (f x y))
                                 _           -> Nothing

      onlyIntegerBinRel :: (Integer -> Integer -> Bool) -> [Value] -> Maybe Value
      onlyIntegerBinRel f vs = case mapM onlyInteger vs of
                                 Just [x, y] -> Just (ValueBool (f x y))
                                 _           -> Nothing

      toDoubleBinOp     :: (Double -> Double -> Double) -> [Value] -> Maybe Value
      toDoubleBinOp     f vs = case mapM toDouble vs of
                                 Just [x, y] -> Just (ValueDouble (f x y))
                                 _           -> Nothing

      toDoubleBinRel    :: (Double -> Double -> Bool) -> [Value] -> Maybe Value
      toDoubleBinRel    f vs = case mapM toDouble vs of
                                 Just [x, y] -> Just (ValueBool (f x y))
                                 _           -> Nothing

      onlyIntegerUnOp   :: (Integer -> Integer) -> [Value] -> Maybe Value
      onlyIntegerUnOp   f vs = case mapM onlyInteger vs of
                                 Just [x] -> Just (ValueInteger (f x))
                                 _        -> Nothing

      toDoubleUnOp      :: (Double -> Double) -> [Value] -> Maybe Value
      toDoubleUnOp      f vs = case mapM toDouble vs of
                                 Just [x] -> Just (ValueDouble (f x))
                                 _        -> Nothing

      coerceItoDOp      :: (Integer -> Double) -> [Value] -> Maybe Value
      coerceItoDOp      f vs = case mapM onlyInteger vs of
                                 Just [x] -> Just (ValueDouble (f x))
                                 _        -> Nothing

      coerceDtoIOp      :: (Double -> Integer) -> [Value] -> Maybe Value
      coerceDtoIOp      f vs = case mapM toDouble vs of
                                 Just [x] -> Just (ValueInteger (f x))
                                 _        -> Nothing

      stringBinRel      :: (String -> String -> Bool) -> [Value] -> Maybe Value
      stringBinRel      f vs = case mapM onlyString vs of
                                 Just [x, y] -> Just (ValueBool (f x y))
                                 _           -> Nothing

      firstJust :: [Maybe a] -> Maybe a
      firstJust [] = Nothing
      firstJust (Nothing : xs) = firstJust xs
      firstJust (Just x  : _ ) = Just x

      typeCasesOp :: [[Value] -> Maybe Value] -> [Value] -> Maybe Value
      typeCasesOp ts v = firstJust (map (\f -> f v) ts)

      makeBuiltin :: String -> Int -> ([Value] -> Maybe Value) -> (String, Value)
      makeBuiltin name arity f = (name, ValueBuiltin { valueBuiltin = do args <- sequence (replicate arity pop)
                                                                         case f (reverse args) of
                                                                           Just v -> push v
                                                                           Nothing -> throwError (BuiltinTypeError name),
                                                       valueBuiltinName = name })

      makeBinMath :: (forall a. (Num a) => a -> a -> a) -> String -> (String, Value)
      makeBinMath f n = makeBuiltin n 2 (typeCasesOp [onlyIntegerBinOp f, toDoubleBinOp f])

      makeUnMath :: (forall a. (Num a) => a -> a) -> String -> (String, Value)
      makeUnMath f n = makeBuiltin n 1 (typeCasesOp [onlyIntegerUnOp f, toDoubleUnOp f])

      makeUnDouble :: (Double -> Double) -> String -> (String, Value)
      makeUnDouble f n = makeBuiltin n 1 (toDoubleUnOp f)

      makeItoD :: (Integer -> Double) -> String -> (String, Value)
      makeItoD f n = makeBuiltin n 1 (coerceItoDOp f)

      makeDtoI :: (Double -> Integer) -> String -> (String, Value)
      makeDtoI f n = makeBuiltin n 1 (coerceDtoIOp f)

      makeBinRel :: (forall a. (Eq a, Ord a) => a -> a -> Bool) -> String -> (String, Value)
      makeBinRel f n = makeBuiltin n 2 (typeCasesOp [onlyIntegerBinRel f, toDoubleBinRel f, stringBinRel f])

      genEq :: [Value] -> Maybe Value
      genEq [x, y] = Just (ValueBool (x == y))
      genEq _      = Nothing

      ifOp :: [Value] -> Maybe Value
      ifOp [ValueBool True,  t, _] = Just t
      ifOp [ValueBool False, _, f] = Just f
      ifOp _                       = Nothing

      consOp :: [Value] -> Maybe Value
      consOp [x, y] = Just (ValuePair x y)
      consOp _      = Nothing

      pairOp :: ((Value, Value) -> Value) -> [Value] -> Maybe Value
      pairOp f [ValuePair x y] = Just (f (x, y))
      pairOp _ _               = Nothing

      errorBuiltin = ValueBuiltin { valueBuiltin = do x <- pop
                                                      case x of
                                                        (ValueString s) -> throwError (UserError s)
                                                        x               -> throwError (BuiltinTypeError "error"),
                                    valueBuiltinName = "error" }
