-- | Haskell types related to Gulik.
module Gulik.Types(-- * Basic types
                   Expr(..),
                   Env,
                   Value(..),
                   -- * Interp monad and error handling
                   Interp,
                   InterpState(..),
                   InterpError(..)) where
-- Import a bunch of things, with care towards name collisions.
import Data.Map as Map hiding(map)
import Control.Monad.State
import Control.Monad.Error

-- | Algebraic data type for Gulik expressions.
data Expr =   ExprDouble  { exprDouble  :: Double  } -- ^ A double-valued literal.
            | ExprInteger { exprInteger :: Integer } -- ^ An integer-valued literal.
            | ExprString  { exprString  :: String  } -- ^ A string-valued literal.
            | ExprApply                              -- ^ An application.
            | ExprName    { exprName    :: String  } -- ^ A name.
            | ExprBind    { exprBind    :: String  } -- ^ A bind to a name.
            | ExprBody    { exprBody    :: [Expr]  } -- ^ A body (list of expressions in {...}).
                     deriving Eq

instance Show Expr where
    show (ExprDouble  d) = show d
    show (ExprInteger i) = show i
    show (ExprString  s) = s
    show ExprApply       = "apply"
    show (ExprName n   ) = n
    show (ExprBind n   ) = '/' : n
    show (ExprBody b   ) = "{" ++ (unwords (map show b)) ++ "}"

-- | An environment, mapping names to values.
type Env = Map String Value

{- Values are either a basic data type (double, bool, string), a
   function closure, or a "builtin closure" (which is first-class like
   a regular closure, but whose body is a transformer on the stack
   machine state written directly in Haskell rather than a list of
   expressions). -}
-- | Algebraic data type for Gulik run-time values.
data Value = -- | A double value
               ValueDouble  { valueDouble :: Double    {- ^ Extract the double.  -} }
             -- | An integer value.
             | ValueInteger { valueInteger :: Integer  {- ^ Extract the integer. -} }
             -- | A Boolean value.
             | ValueBool    { valueBool :: Bool        {- ^ Extract the Boolean. -} }
             -- | A string value.
             | ValueString  { valueString :: String    {- ^ Extract the string.  -} }
             -- | A nil pair.
             | ValueNil
             -- | A (non-nil) pair.
             | ValuePair    { valueCar :: Value,          {- ^ The car of the pair. -}
                              valueCdr :: Value           {- ^ The cdr of the pair. -} }
             -- | A function closure, written in Gulik.
             | ValueClosure { valueEnv :: Env,            {- ^ The environment for the closure. -}
                              valueBody :: [Expr]         {- ^ The body of the closure.         -} }
             -- | A function closure, written in Haskell.
             | ValueBuiltin { valueBuiltin :: Interp (),  {- ^ The Interp monad action for the builtin.               -}
                              valueBuiltinName :: String  {- ^ The name of the builtin (used for superficial things). -} }

{- Equality is defined in a straightforward way; in particular, two
   builtins are equal iff they are assigned the same valueBuiltinName
   (which is tracked independently of possible environment bindings,
   and in fact is only used for equality and printing). -}
instance Eq Value where
    ValueDouble x    == ValueDouble y    = x == y
    ValueInteger x   == ValueInteger y   = x == y
    ValueBool x      == ValueBool y      = x == y
    ValueString x    == ValueString y    = x == y
    ValueNil         == ValueNil         = True
    ValuePair x xs   == ValuePair y ys   = ((x == y) && (xs == ys))
    ValueBuiltin _ x == ValueBuiltin _ y = x == y
    ValueClosure e x == ValueClosure f y = ((e == f) && (x == y))
    _                == _                = False

-- Show a value without constructor names.
instance Show Value where
    show (ValueDouble x)    = show x
    show (ValueInteger x)   = show x
    show (ValueBool x)      = show x
    show (ValueString x)    = show x
    show ValueNil           = "--nil--"
    show (ValuePair x xs)   = show (x,xs)
    show (ValueClosure _ _) = "--closure--"
    show (ValueBuiltin _ n) = "--builtin " ++ n ++ "--"

-- | State of the interpreter stack machine.
data InterpState = InterpState { stateStack :: [Value],  {- ^ Current stack.       -}
                                 stateEnv :: Env         {- ^ Current environment. -} }
                   deriving (Eq, Show)

-- | Errors thrown by the stack machine.
data InterpError = -- | A symbol was accessed before being defined.
                     UndefinedSymbol  { errorSymbolName  :: String  {- ^ Name of the undefined symbol. -}        }

                   -- | A pop was executed on an empty stack.
                   | StackUnderflow

                   -- | A non-closure was applied.
                   | InvalidApply     { errorApplyValue  :: Value   {- ^ Value for which apply was attempted. -} }

                   -- | A builtin function was applied with bad types.
                   | BuiltinTypeError { errorBuiltinName :: String  {- ^ The name of the builtin applied. -}     }

                   -- | A user-defined error occurred.
                   | UserError        { errorUserString  :: String  {- ^ A string describing the error. -}       }

                   -- | An unknown error occurred.
                   | UnknownError

instance Error InterpError where
    noMsg    = UnknownError
    strMsg s = UserError { errorUserString = s }

instance Show InterpError where
    show (UndefinedSymbol s)  = "Undefined symbol " ++ s
    show StackUnderflow       = "Stack underflow"
    show (InvalidApply v)     = "Attempt to apply non-closure " ++ (show v)
    show (BuiltinTypeError n) = "Bad types for builtin " ++ n
    show (UserError s)        = "User error: " ++ s
    show UnknownError         = "Unknown error"

-- | Monad for the stack machine operations, combining state and error handling.
type Interp = StateT InterpState (Either InterpError)
